
# As it is now, selects only the best analog
Analogs <- function(exp, obs, nanalogs = 3, fun_analog = "mean", ncores = 1) {

  require(multiApply)
  require(CSTools)

  #-----------------------------------
  # Checkings
  #-----------------------------------
  # input exp and obs must be s2dv_cube objects
  if (!inherits(exp,'s2dv_cube')) {
    stop("**ERROR: Parameter 'exp' must be of the class 's2dv_cube'")
  }

  # input exp and obs must be s2dv_cube objects
  if (!inherits(obs,'s2dv_cube')) {
    stop("**ERROR: Parameter 'obs' must be of the class 's2dv_cube'")
  }

  # Select a function to apply to the analogs selected for a given observation
  stopifnot(fun_analog %in% c("mean", "wmean", "max", "min", "median"))

  # sdate must be the time dimension in the input data
  #stopifnot('sdate' %in% names(dim(exp$data)))
  #stopifnot('sdate' %in% names(dim(obs$data)))

  #-----------------------------------
  # Interpolate high-res observations to the coarse grid
  #-----------------------------------
  grid_exp <- exp$source_files[1]  
  
  # Since we are interpolated to a coarser grid, round bordering coordinates
  attr(obs$lon,"first_lon") <- round(attr(obs$lon,"first_lon"))
  attr(obs$lon,"last_lon") <- round(attr(obs$lon,"last_lon"))
  attr(obs$lat,"first_lat") <- round(attr(obs$lat,"first_lat"))
  attr(obs$lat,"last_lat") <- round(attr(obs$lat,"last_lat"))
 
  obs_interpolated <- Interpolation(exp = obs, target_grid = grid_exp, method_remap = "conservative")
  lats <- obs$lat
  lons <- obs$lon

  #-----------------------------------
  # Reshape train and test
  #-----------------------------------
  res <- Apply(list(obs_interpolated$data, exp$data, obs$data), target_dims = list(c('sdate', 'lat', 'lon'), 
               c('sdate', 'lat', 'lon'), c('sdate', 'lat', 'lon')), fun = function(tr, te, ob) 
               .analogs(train = tr, test = te, obs_hres = ob, k = nanalogs, fun = fun_analog))$output1

  res_s2dv <- suppressWarnings(s2dv_cube(data = res, lon = lons, lat = lats))
  
  return(res_s2dv)
}


# For each element in test return the corresponding k-NN observations, or its mean if indicated in fun
.analogs <- function(train, test, obs_hres, k, fun = NULL) {

  require(FNN)
  require(multiApply)

  # Get knn indices for each test element
  knn.ind <- get_knn(train = train, test = test, k = k)  

  dist <- knn.ind$nn.dist
  idx <- knn.ind$nn.index
  names(dim(idx)) <- c("sdate", "analog"); names(dim(dist)) <- c("sdate", "analog")

  # Obtain the obs for the knn indices
  analogs <- Apply(idx, margins = "analog", fun = function(an) { obs_hres[an, , ] })$output1

  # Apply function for analogs
  if(!is.null(fun)) {
  
    margins_analogs <- names(dim(analogs))[names(dim(analogs)) != "analog"]
   
    if (fun == "wmean") {
      w <- 1 / dist 
      analogs <- Apply(list(analogs, dist), margins = list(margins_analogs, "sdate"), 
                       fun = function(x, w) weighted.mean(x, w))$output1
    } else {   
      analogs <- Apply(analogs, margins = margins_analogs, fun = fun)$output1
    }
  }

  return(analogs)
}

# For each element in test, find the indices of the k nearest neigbhors in train
get_knn <- function(train, test, k) {
  # Reformat train and test as an array with (time, points)
  train <- aperm(Apply(train, target_dims = list(c('lat', 'lon')), fun=function(x) { as.vector(x) })$output1)
  test <- aperm(Apply(test, target_dims = list(c('lat', 'lon')), fun=function(x) { as.vector(x) })$output1)

  # For each row in test, find the k closest rows in train
  knn.ind <- get.knnx(train, test, k)
  return(knn.ind)
}


