#method_remap implemented:  con, bil, bic, nn, con2
#method_point_interp: nearest, bilinear, 9point, invdist4nn, NE, NW, SE, SW
#target_grid has to be either a grid recognised by CDO or a NETCDF file
# requires CDO_1.9.8 or newer versions when using nn method
Interpolation <- function(exp, points = NULL, method_remap = NULL, target_grid = NULL,
                          method_point_interp = NULL) 
{
  require(s2dv)

  # Input data must be an s2dv_cube object
  if (!inherits(exp,'s2dv_cube')) {
    stop("**ERROR: Parameter 'exp' must be of the class 's2dv_cube'")
  }

  # checkings for the case of interpolation to point locations
  if (!is.null(points)) {
    if (!inherits(points, 'list')) {
      stop("**ERROR: Parameter 'points' must be a list of two elements containing the point ",
           "latitudes and longitudes in the form 'list(c(latitudes), c(longitudes))'")
    }
 
    if (is.null(method_point_interp)) {
      stop("**ERROR: Parameter 'method_point_interp' must be a character vector indicating the ",
           "interpolation method. Accepted methods are nearest, bilinear, 9point, ",
           "invdist4nn, NE, NW, SE, SW")
    }
 
    # Points must be a list of two elements
    if (length(points) != 2) {
      stop("**ERROR: 'points' must be a lis of two elements containing the point ",
           "latitudes and longitudes in the form 'list(c(latitudes), c(longitudes))'")
    }

    # Check that the number of latitudes and longitudes match
    if (length(unique(lengths(points))) != 1L) {
      stop("**ERROR: the number of latitudes and longitudes must match")
    }
  } else {
    if (is.null(method_remap)) {
      stop("**ERROR: Parameter 'method_remap' must be a character vector indicating the ",
           "interpolation method. Accepted methods are con, bil, bic, nn, con2")
    }
    if (is.null(target_grid)) {
      stop("**ERROR: Parameter 'target_grid' must be a character vector indicating the ",
           "target grid to be passed to CDO. It must be a grid recognised by CDO ",
           "or a NetCDF file")  
    }
  }

  lonmin <- attr(exp$lon,"first_lon")
  lonmax <- attr(exp$lon,"last_lon")
  latmin <- attr(exp$lat,"first_lat")
  latmax <- attr(exp$lat,"last_lat")

  #----------------------------------
  # Map regrid with CDO
  #----------------------------------
  if (is.null(points)) {
    res <- CDORemap(data_array = exp$data, 
                    lats = exp$lat, 
                    lons = exp$lon, 
                    grid = target_grid, 
                    method = method_remap, 
                    crop = c(lonmin, lonmax, latmin, latmax))

    # Create an s2dv_cube object
    res_s2dv <- suppressWarnings(s2dv_cube(data = res$data_array, lon = res$lons, lat = res$lats))

  #----------------------------------
  # Interpolate to point locations
  #----------------------------------
  } else {  
    # First create interpolation weights, depending on the chosen method
    ncfile_exp <- exp$source_files[1]  
    weights <- create_interp_weights(ncfile = ncfile_exp, locids = 1:unique(lengths(points)), 
                                     lats = points[[1]], lons = points[[2]], method = method_point_interp) 

    # Select coarse-scale data to be interpolated
    model_data_gridpoints <- get_model_data(weights.df = weights, mdata = exp$data)

    # Interpolate model data to point locations
    res <- interpolate_data(model_data_gridpoints, weights)
  
    # Create an s2dv_cube object
    res_s2dv <- suppressWarnings(s2dv_cube(data = res, lon = points[[1]], lat = points[[2]])) 
                               
  }

  return(res_s2dv)
}


#======================
# Compute weights for interpolation at several (lat,lon) positions
# We assume that grid boxes are centered in the grid point.
#======================
create_interp_weights <- function(ncfile, locids, lats, lons, 
                                  method = c("nearest", "bilinear", "9point", "invdist4nn", "NE", 
                                             "NW", "SE", "SW")) 
{ #----------------
  # Read grid description and compute (i,j) of requested locations (including decimals)
  #----------------
  griddes <- get_griddes(ncfile)
  if (!is.null(griddes)) {
    stop("**ERROR: 'griddes' not found in the NetCDF source files")
  } 
  gridpoints <- latlon2ij(griddes, lats, lons)

  #----------------
  # Compute the weights according to the selected method
  #----------------
  if(method == "nearest") {
    #----------------
    # Round i and j to closest integer. Weight is always 1.
    #----------------

    #  |     |     |
    # -+-----+-----+-
    #  |    x|     |
    #  |  a  |     |
    #  |     |     |
    # -+-----+-----+-
    #  |     |     |

    centeri <- round(gridpoints$i,0)
    centeri[centeri == griddes$xsize+1] <- 1 # close longitudes		

    weights.df <- data.frame(locid = locids,
                             lat = lats,
                             lon = lons,
                             rawi = gridpoints$i,
                             rawj = gridpoints$j,
                             i = centeri,
                             j = round(gridpoints$j, 0),
                             weight = 1)
  } else if (method %in% c("bilinear","invdist4nn")) {
    #----------------
    # Get the (i,j) coordinates of the 4 points (a,b,c,d) around x.
    # This plot shows i increasing to the right and 
    # j increasing to the top, but the computations are generic.
    #----------------
    #  |     |     |
    #- +-----+-----+-
    #  |     |     |
    #  |  b  |  c  |
    #  |     |     |
    #- +-----+-----+-
    #  |    x|     |
    #  |  a  |  d  |
    #  |     |     |
    #- +-----+-----+-
    #  |     |     |

    lowi <- floor(gridpoints$i)
    highi <- ceiling(gridpoints$i)
    highi[highi == griddes$xsize+1] <- 1  # close the longitudes
    lowj <- floor(gridpoints$j)
    highj <- ceiling(gridpoints$j)
    # Note: highi and lowi are the same if i is integer
    # Note: highj and lowj are the same if j is integer

    #----------------
    # Get x position wrt ad and ab axes (from 0 to 1)
    #----------------
    pcti <- gridpoints$i - lowi
    pctj <- gridpoints$j - lowj

    #----------------
    # Compute weights for a,b,c,d grid points
    #----------------
    if(method == "bilinear") {
      wa = (1 - pcti) * (1 - pctj)
      wb = (1 - pcti) * pctj
      wc = pcti * pctj
      wd = pcti * (1 - pctj)
    } else if(method == "invdist4nn") {
      #----------------
      # Note: the distance is computed in the (i,j) space.
      # Note2: this method does not guarantees a continuous interpolation.
      # Use bilinear if that's desirable. 
      # When x is on the ab line, c and d would be used. In the limit of x 
      # being just left of ab other points would be used.
      # Here we just dropped c and d coeffs when over ab. Same for ad line, 
      # b and c coeffs dropped. This prevents repeated nodes.
      #----------------
      ida = 1 / sqrt(pcti^2 + pctj^2)
      idb = 1 / sqrt(pcti^2 + (1 - pctj)^2)
      idc = 1 / sqrt((1-pcti)^2 + (1-pctj)^2)
      idd = 1 / sqrt((1-pcti)^2 + pctj^2)
      idb[pctj == 0] <- 0; 
      idc[pctj == 0] <- 0; 
      idc[pcti == 0] <- 0;
      idd[pcti == 0] <- 0;
 
      #----------------
      # Normalize vector of inverse distances
      #----------------
      invdist <- cbind(ida, idb, idc, idd)
      print(invdist)
      w <- t(apply(invdistc(1),function(x) { print(x); if(any(is.infinite(x))) {
                   x <- is.infinite(x) * 1 } ; x <- x/sum(x) }))
      print(w)

      wa = w[ , 1]
      wb = w[ , 2]
      wc = w[ , 3]
      wd = w[ , 4]
    }

    #----------------
    # Put info in dataframes and rbind them
    #----------------
    weightsa.df <- data.frame(locid = locids, lat = lats,lon = lons, rawi = gridpoints$i,
                              rawj = gridpoints$j, i = lowi, j = lowj, weight = wa)
    weightsb.df <- data.frame(locid = locids, lat = lats, lon = lons, rawi = gridpoints$i,
                              rawj = gridpoints$j, i = lowi, j = highj, weight = wb)
    weightsc.df <- data.frame(locid = locids, lat = lats, lon = lons, rawi = gridpoints$i,
                              rawj = gridpoints$j, i = highi, j = highj, weight = wc)
    weightsd.df <- data.frame(locid = locids, lat = lats, lon = lons, rawi = gridpoints$i,
                              rawj = gridpoints$j, i = highi, j = lowj, weight = wd)
    weights.df <- rbind(weightsa.df, weightsb.df, weightsc.df, weightsd.df)
  } else if(method == "9point") {
    #----------------
    # Get the (i,j) coordinates of the 9 points (a,b,...,i) around x
    # This plot shows i increasing to the right and 
    # j increasing to the top, but the computations are generic.
    #----------------
    # |     |     |     |
    #-+-----+-----+-----+-
    # |     |     |     |
    # |  c  |  f  |  i  |
    # |     |     |     |
    #-+-----+-----+-----+-
    # |     |    x|     |
    # |  b  |  e  |  h  |
    # |     |     |     |
    #-+-----+-----+-----+-
    # |     |     |     |
    # |  a  |  d  |  g  |
    # |     |     |     |
    #-+-----+-----+-----+-
    # |     |     |     |

    centeri <- round(gridpoints$i, 0)
    centeri[centeri == griddes$xsize + 1] <- 1 
    centerj <- round(gridpoints$j, 0)
    lowi <- centeri - 1
    highi <- centeri + 1
    lowi[lowi == 0] <- griddes$xsize # close the longitudes
    highi[highi == griddes$xsize+1] <- 1 # close the longitudes
    lowj <- centerj - 1
    highj <- centerj + 1
		
    #----------------
    # For the north and south pole do a 6-point average
    #----------------
    w_highj   <- ifelse(centerj == 1,1/6,ifelse(centerj == griddes$ysize,0  ,1/9))
    w_centerj <- ifelse(centerj == 1,1/6,ifelse(centerj == griddes$ysize,1/6,1/9))
    w_lowj    <- ifelse(centerj == 1,0  ,ifelse(centerj == griddes$ysize,1/6,1/9))

    #----------------
    # Put info in dataframes and rbind them
    #----------------
    weightsa.df <- data.frame(locid = locids, lat = lats, lon = lons, rawi = gridpoints$i,
                              rawj = gridpoints$j, i = lowi, j = lowj, weight = w_lowj)
    weightsb.df <- data.frame(locid = locids, lat = lats, lon = lons, rawi = gridpoints$i,
                              rawj = gridpoints$j, i = lowi, j = centerj, weight = w_centerj)
    weightsc.df <- data.frame(locid = locids, lat = lats, lon = lons, rawi = gridpoints$i,
                              rawj = gridpoints$j, i = lowi, j = highj, weight = w_highj)
    weightsd.df <- data.frame(locid = locids, lat = lats, lon = lons, rawi = gridpoints$i,
                              rawj = gridpoints$j, i = centeri, j = lowj, weight = w_lowj)
    weightse.df <- data.frame(locid = locids, lat = lats, lon = lons, rawi = gridpoints$i,
                              rawj = gridpoints$j, i = centeri, j = centerj, weight = w_centerj)
    weightsf.df <- data.frame(locid = locids, lat = lats, lon = lons, rawi = gridpoints$i,
                              rawj = gridpoints$j, i = centeri, j = highj, weight = w_highj)
    weightsg.df <- data.frame(locid = locids, lat = lats, lon = lons, rawi = gridpoints$i,
                              rawj = gridpoints$j, i = highi, j = lowj, weight = w_lowj)
    weightsh.df <- data.frame(locid = locids, lat = lats, lon = lons, rawi = gridpoints$i,
                              rawj = gridpoints$j, i = highi, j = centerj, weight = w_centerj)
    weightsi.df <- data.frame(locid = locids, lat = lats, lon = lons, rawi = gridpoints$i,
                              rawj = gridpoints$j, i = highi, j = highj, weight = w_highj)
    weights.df <- rbind(weightsa.df, weightsb.df, weightsc.df, weightsd.df, weightse.df,
                        weightsf.df, weightsg.df, weightsh.df, weightsi.df)
  } else if(method %in% c("NE", "NW", "SW", "SE")) {
    #----------------
    # Find if increasing i and j increases or decreases lat and lon
    #----------------
    westtoeast <- (griddes$xinc > 0)
    southtonorth <- T
    if(griddes$gridtype == "gaussian") { 
      # We assume gaussian grid latitudes are ordered north to south
      southtonorth <- F 
      } else { #lonlat
        if(griddes$yinc < 0) {southtonorth <- F}
      }

      #----------------
      # Get the (i,j) coordinates of the desired point (a,b,c or d) around x
      #----------------
      #  |     |     |
      #- +-----+-----+-
      #  |     |     |
      #  |  b  |  c  |
      #  |     |     |
      #- +-----+-----+-
      #  |    x|     |
      #  |  a  |  d  |
      #  |     |     |
      #- +-----+-----+-
      #  |     |     |

      if(substr(method,1,1) == "N" & southtonorth == T) { selj <- ceiling(gridpoints$j) }
      if(substr(method,1,1) == "S" & southtonorth == T) { selj <- floor(gridpoints$j) }
      if(substr(method,1,1) == "N" & southtonorth == F) { selj <- floor(gridpoints$j) }
      if(substr(method,1,1) == "S" & southtonorth == F) { selj <- ceiling(gridpoints$j) }

      if(substr(method,2,2) == "E" & westtoeast == T) {seli <- ceiling(gridpoints$i) }
      if(substr(method,2,2) == "W" & westtoeast == T) {seli <- floor(gridpoints$i) }
      if(substr(method,2,2) == "E" & westtoeast == F) {seli <- floor(gridpoints$i) }
      if(substr(method,2,2) == "W" & westtoeast == F) {seli <- ceil(gridpoints$i) }
      
      seli[seli == griddes$xsize + 1] <- 1  # close the longitudes
      
      weights.df <- data.frame(locid = locids, lat = lats, lon = lons, rawi = gridpoints$i, 
                               rawj = gridpoints$j, i = seli, j = selj, weight = 1)
  } else { 
    stop(paste0("**ERROR: Method " ,method, " not implemented"))
  }
  
  #----------------  
  # Order by locid and remove lines with 0 weight 
  # This also removes some duplicates in the bilinear/invdist4nn methods when i 
  # or j is a whole number, or in the 9-point method when at the poles.
  #----------------
  weights.df <- weights.df[order(weights.df$locid), ]
  weights.df <- weights.df[weights.df$weight != 0, ]
	
  #----------------
  # Add as attributes the method and the nc file used to compute the weights
  #----------------
  attributes(weights.df)$nc_file <- normalizePath(ncfile)
  attributes(weights.df)$method <- method

  return(weights.df)
}

#======================
# Compute (i,j) from (lat,lon).
# Works only for 'lonlat' and 'gaussian' grids.
# Grids are supposed to cover whole globe.
#======================
latlon2ij <- function(griddes, lats, lons) {
  #------------
  # Check input params
  #------------
  if(length(lons) != length(lats)) {stop("**ERROR: input lat and lon have different lengths.")}
  if(any(lats < -90) | any(lats > 90)) {stop("**ERROR: latitude out of valid range")}
  if(round(griddes$xinc*griddes$xsize)!=360) {stop("**ERROR: grid is not global")}
  # no need to resize lons to [0,360)

  #------------
  # Compute i (with decimals)
  # i lies in [1,xsize+1)
  #------------
  gridpoints <- list()
  gridpoints$i <- 1 + (((lons-griddes$xfirst) / griddes$xinc) %% griddes$xsize)

  #------------
  # Compute j (with decimals)
  #------------
  if(griddes$gridtype=='lonlat') {
    gridpoints$j <- 1 + (lats-griddes$yfirst) / griddes$yinc 
  } else if(griddes$gridtype == 'gaussian') {
    # We assume gaussian grid latitudes are ordered north to south
    # findInterval can only work with monotonic ascending values so we revert twice
    northj <- griddes$ysize-findInterval(lats, -griddes$yvals)
    southj <- northj + 1

    # Special case: We are north of the first lat
    gridpoints$j[northj == 0] <- 1

    # Special case: We are south of the last lat
    gridpoints$j[southj == griddes$ysize + 1] <- griddes$ysize

    # Generic case
    ok_idx <- !(northj == 0 | southj == griddes$ysize+1)
    gridpoints$j[ok_idx] <- northj[ok_idx] + (griddes$yvals[northj[ok_idx]] - 
                            lats[ok_idx])/(griddes$yvals[northj[ok_idx]] - griddes$yvals[southj[ok_idx]])
  } else { stop("**ERROR: unsupported grid") }
	
  return(gridpoints)
}

#======================
# Use cdo griddes to obtain grid information
#======================
get_griddes <- function(ncfile) {
  tmp <- system(paste0("cdo griddes ", ncfile,
                       " 2>/dev/null | egrep 'gridtype|xsize|ysize|xfirst|xinc|yfirst|yinc'"), intern = T)
  arr <- do.call(rbind, strsplit(tmp,"\\s+= ", perl = T))
  griddes <- as.list(arr[,2])
  names(griddes) <- arr[,1]

  if(griddes$gridtype == "gaussian") {
    griddes$yvals <- get_lats(ncfile)
  }

  # Convert some fields to numeric. Ensures all fields are present.
  for(nm in c("xsize", "ysize", "xfirst", "yfirst", "xinc", "yinc")) {
    griddes[[nm]] <- ifelse(is.null(griddes[[nm]]), NA, as.numeric(griddes[[nm]]))
  }
	
  return(griddes)
}

#======================
# Use nco to obtain latitudes. Latitudes shall be named "lat" or "latitude".
#======================
get_lats <- function(ncfile) {

  tmp <- system(paste0('ncks -H -s "%f " -v latitude ',ncfile),intern=T)

  if(!is.null(attributes(tmp)$status)) {
    tmp <- system(paste0('ncks -H -s "%f " -v lat ',ncfile),intern=T)
  }
  
  lats <- as.numeric(strsplit(tmp[1],"\\s+",perl=T)[[1]])

  return(lats)
}

#======================
# Load model data at all (i,j) pairs listed in the weights dataframe. 
# Uses StartR. All ... parameters go to Start (i.e. specify dat, var, 
# sdate, time, ensemble, num_procs, etc)
#======================
get_model_data <- function(weights.df, mdata) {
 
  require(plyr)

  #-----------------
  # Get data for all combinations of i and j.
  # (inefficient, getting many unneded pairs).
  # Avoid retrieving duplicates with unique()
  #-----------------
  is <- weights.df$i
  js <- weights.df$j
 
  #-----------------
  # Get indices of original is and js in unique(is),unique(js) that were requested
  #-----------------
  idxi <- match(is, unique(is))
  idxj <- match(js, unique(js))

  #-----------------
  # Subsample mdata to keep only the needed (i,j) pairs.
  #-----------------
  if (is.na(match("longitude", names(dim(mdata))))) {
    londim <- match("lon", names(dim(mdata)))
  } else {
    londim <- match("longitude", names(dim(mdata)))
  }
  if (is.na(match("latitude", names(dim(mdata))))) {
    latdim <- match("lat", names(dim(mdata)))
  } else {
    latdim <- match("latitude", names(dim(mdata)))
  }

  # trick: exchange idxi and idxj
  if(londim > latdim) { idx.tmp <- idxi; idxi <- idxj; idxj <- idx.tmp }
  keepdims <- (1:length(dim(mdata)))[-c(londim,latdim)]

  sub_mdata <- apply(mdata,keepdims, function(x) { 
                     laply(1:length(is),function(k) { x[idxi[k],idxj[k]] }) })
  names(dim(sub_mdata))[1] <- "gridpoint"

  #-----------------
  # Return an array that contains as many gridpoints as (i,j) pairs were requested
  #-----------------
  return(sub_mdata)
}

#======================
# Multiply the grid-point series by the weights, 
# to obtain the desired interpolations
#======================
interpolate_data <- function(model_data, weights.df) {
  #-----------------
  # Multiply each gridpoint matrix by its corresponding weight
  #-----------------
  gpdim <- match("gridpoint", names(dim(model_data)))
  weighted_data <- sweep(model_data, gpdim, weights.df$weight, "*")

  #-----------------
  # Sum all series that belong to same interpolation point
  # Return an array that contains the requested locations and interpolation type
  #-----------------
  interp_data <- apply(weighted_data, -gpdim, function(x) { rowsum(x, weights.df$locid) }) 
  names(dim(interp_data))[1] <- "location"
  return(interp_data)
}


